@@ -31,21 +31,27 @@
 # @brief Main module for note-python. Contains core library functionality.
 
 import sys
+import os
 import json
 import time
 
 use_periphery = False
 use_micropython = False
 use_serial_lock = False
+use_circuitpython = sys.implementation.name == 'circuitpython'
+use_micropython = sys.implementation.name == 'micropython'
+use_rtc = not use_micropython and not use_circuitpython
+
 if sys.implementation.name == 'cpython':
-    if sys.platform == "linux" or sys.platform == "linux2":
+    if sys.platform == 'linux' or sys.platform == 'linux2':
         use_periphery = True
         from periphery import I2C
 
         use_serial_lock = True
         from filelock import Timeout, FileLock
-elif sys.implementation.name == 'micropython':
-    use_micropython = True
+
+use_i2c_lock = not use_periphery and not use_micropython
+
 
 NOTECARD_I2C_ADDRESS = 0x17
 
@@ -55,6 +61,33 @@ NOTECARD_I2C_ADDRESS = 0x17
 CARD_REQUEST_SEGMENT_MAX_LEN = 250
 CARD_REQUEST_SEGMENT_DELAY_MS = 250
 
+if not use_rtc:
+    if use_circuitpython:
+        import supervisor
+        from supervisor import ticks_ms
+
+        def ticks_diff(ticks1, ticks2):
+            """Compute the signed difference between two ticks values."""
+            diff = (ticks1 - ticks2) & _TICKS_MAX  # noqa: F821
+            diff = ((diff + _TICKS_HALFPERIOD)  # noqa: F821
+                    & _TICKS_MAX) - _TICKS_HALFPERIOD  # noqa: F821
+            return diff
+    if use_micropython:
+        from utime import ticks_diff, ticks_ms  # noqa: F811
+
+
+def has_timed_out(start, timeout_secs):
+    """Determine whether a timeout interval has passed during communication."""
+    if not use_rtc:
+        return ticks_diff(ticks_ms(), start) > timeout_secs * 1000
+    else:
+        return time.time() > start + timeout_secs
+
+
+def start_timeout():
+    """Start the timeout interval for I2C communication."""
+    return ticks_ms() if not use_rtc else time.time()
+
 
 def prepareRequest(req, debug=False):
     """Format the request string as a JSON object and add a newline."""
@@ -155,16 +188,49 @@ class Notecard:
     to reset the Notecard via Serial or I2C.
     """
 
+    _user_agent = {
+        'agent': 'note-python',
+        'os_name': sys.implementation.name,
+        'os_platform': sys.platform,
+        'os_version': sys.version
+    }
+    _user_agent_sent = False
+
     def __init__(self):
-        """Initialize the Notecard through a reset."""
+        """Initialize the Notecard through a reset and configure user agent."""
         self.Reset()
 
+        if sys.implementation.name == 'cpython':
+            self._user_agent['os_family'] = os.name
+        else:
+            self._user_agent['os_family'] = os.uname().machine
+
+    def _preprocessReq(self, req):
+        """Inspect the request for hub.set and add the User Agent."""
+        if 'hub.set' in req.values():
+            # Merge the User Agent to send along with the hub.set request.
+            new_req = req.copy()
+            new_req.update(self.GetUserAgent())
+            req = new_req
+
+            self._user_agent_sent = True
+        return req
+
+    def GetUserAgent(self):
+        """Return the User Agent String for the host for debug purposes."""
+        return self._user_agent
+
+    def UserAgentSent(self):
+        """Return true if the User Agent has been sent to the Notecard."""
+        return self._user_agent_sent
+
 
 class OpenSerial(Notecard):
     """Notecard class for Serial communication."""
 
     def Command(self, req):
         """Perform a Notecard command and exit with no response."""
+        req = self._preprocessReq(req)
         if 'cmd' not in req:
             raise Exception("Please use 'cmd' instead of 'req'")
 
@@ -181,6 +247,7 @@ class OpenSerial(Notecard):
 
     def Transaction(self, req):
         """Perform a Notecard transaction and return the result."""
+        req = self._preprocessReq(req)
         if use_serial_lock:
             try:
                 self.lock.acquire(timeout=5)
@@ -207,6 +274,9 @@ class OpenSerial(Notecard):
 
     def __init__(self, uart_id, debug=False):
         """Initialize the Notecard before a reset."""
+        self._user_agent['req_interface'] = 'serial'
+        self._user_agent['req_port'] = str(uart_id)
+
         self.uart = uart_id
         self._debug = debug
 
@@ -245,6 +315,7 @@ class OpenI2C(Notecard):
 
     def Command(self, req):
         """Perform a Notecard command and exit with no response."""
+        req = self._preprocessReq(req)
         if 'cmd' not in req:
             raise Exception("Please use 'cmd' instead of 'req'")
 
@@ -260,6 +331,8 @@ class OpenI2C(Notecard):
 
     def Transaction(self, req):
         """Perform a Notecard transaction and return the result."""
+        req = self._preprocessReq(req)
+
         req_json = prepareRequest(req, self._debug)
         rsp_json = ""
 
@@ -271,7 +344,7 @@ class OpenI2C(Notecard):
 
             chunk_len = 0
             received_newline = False
-            start = time.time()
+            start = start_timeout()
             transaction_timeout_secs = 10
             while True:
                 time.sleep(.001)
@@ -303,7 +376,7 @@ class OpenI2C(Notecard):
                     continue
                 if received_newline:
                     break
-                if (time.time() >= start + transaction_timeout_secs):
+                if (has_timed_out(start, transaction_timeout_secs)):
                     raise Exception("notecard request or response was lost")
                 time.sleep(0.05)
 
@@ -350,18 +423,21 @@ class OpenI2C(Notecard):
 
     def lock(self):
         """Lock the I2C port so the host can interact with the Notecard."""
-        if not use_periphery and not use_micropython:
+        if use_i2c_lock:
             return self.i2c.try_lock()
         return True
 
     def unlock(self):
         """Unlock the I2C port."""
-        if not use_periphery and not use_micropython:
+        if use_i2c_lock:
             return self.i2c.unlock()
         return True
 
     def __init__(self, i2c, address, max_transfer, debug=False):
         """Initialize the Notecard before a reset."""
+        self._user_agent['req_interface'] = 'i2c'
+        self._user_agent['req_port'] = address
+
         self.i2c = i2c
         self._debug = debug
         if address == 0:
